//------------------------------------------------------------------------------------------------------
//
//  DARC-II.c
//
//  DARC-II ToothPIC application 
//
//  (c) FlexiPanel Ltd
//
//-------------------------------------------------------------------------------------------------------

//Note - when next revisit DARC-II, add:
// - linear scaling for numbers
// - Low level commands by (i) FxP message / (ii) text control

#include "..\Toothpick.h"
#include <p18cxxx.h>
#include "DARCIIConfig.h" 
#include "..\HopCodes.h"

// locate DARC Config structure declared in DARCConfig.c
extern rom DARCIIConfig DARCcfg;

// pLocalName and pszPIN declared in ToothPIC.c
extern rom unsigned char pLocalName[] ;    
extern rom unsigned char pszPIN[] ;    

// reserve RAM used by FlexiPanel user interface to ensure linker doesn't allocate it to anything else - maximum UI ram is 0x800
#pragma udata TP_UIRAM000
unsigned char pFxPRAM000B[0xFF];  // one byte allocated in DARCIIDefaultRes.c
#pragma udata TP_UIRAM100
unsigned char pFxPRAM100[0x100];
#pragma udata TP_UIRAM200
unsigned char pFxPRAM200[0x100];
#pragma udata TP_UIRAM300
unsigned char pFxPRAM300[0x100];
#pragma udata TP_UIRAM400
unsigned char pFxPRAM400[0x100];
#pragma udata TP_UIRAM500
unsigned char pFxPRAM500[0x100];
#pragma udata TP_UIRAM600
unsigned char pFxPRAM600[0x100];
#pragma udata TP_UIRAM700
unsigned char pFxPRAM700[0x100];
#pragma udata

// local function prototypes
void ControlChanged( unsigned char CtlID );
void SetDigitalPin( unsigned char DigitalPinID, unsigned char iVal );
void SetListPin( unsigned char DigitalPinID, unsigned short iVal );
void SetNumberPin( unsigned char DigitalPinID, unsigned short iVal );
void RefreshInputs( void );
unsigned char GetPinVal( unsigned char DigitalPinID );
short GetNumberVal( unsigned char PinID );

// DARC-II semaphores
volatile unsigned char PulseClearCountDown;
volatile unsigned char RefeshInputsNow;
volatile unsigned char ResetMePlease;  

// main program 
void main( void )
{
	unsigned char cVal;

	// if no LinkMatik, flash red led rapidly
	while ((ToothpickSemaphores&TPSF_LMTEXISTS)==0)
	{
		LedRed = ~LedRed;
		msDelay(50);
	}

	PulseClearCountDown = 0;
	RefeshInputsNow = 0;
	ResetMePlease = 0;

	FxPCommand( FxPC_Start, 0, 0 );		// Start FlexiPanel service
	AwaitLMTComplete(); 				// Service operating, no client

	// if darc data has been programmed, set name and pin code
	// new name only appears in device discovery after ToothPIC reset
	if (DARCcfg.Initialized==0x55)
	{
		SetBytes( STR_ROM00, (ADD)pLocalName, DARCcfg.ServerName, 0, MAXNAMELENGTHINCZ );
		SetBytes( STR_ROM00, (ADD)pszPIN, DARCcfg.PINCode, 0, MAXPINLENGTHINCZ );
	}

	if (RefeshInputsNow==1 || (RefeshInputsNow==4 && PulseClearCountDown==8) || PulseClearCountDown==4 ||
		RefeshInputsNow==5 || (RefeshInputsNow==3 && PulseClearCountDown==3) || PulseClearCountDown==3 ||
		RefeshInputsNow==7 || (RefeshInputsNow==2 && PulseClearCountDown==2) || PulseClearCountDown==2 ||
		RefeshInputsNow==2 || (RefeshInputsNow==5 && PulseClearCountDown==1) || PulseClearCountDown==17)
	{
		RefeshInputsNow = RefeshInputsNow + 1;
		RefeshInputsNow = RefeshInputsNow - 1;
	}

	// initialize I/O
	// analog channels
	ADCON1 = (ADCON1 & 0xF0) | (0x0F-DARCcfg.nAnalogChannel);

	if (DARCcfg.IsRefAN2)
		VRefNegIsAN2
	else
		VRefNegIsVss

	if (DARCcfg.IsRefAN3)
		VRefPosIsAN3
	else
		VRefNegIsVss

	if (DARCcfg.nAnalogChannel)
		ADConverterOn10bit
	else
		ADConverterOff

	// PWM
	if (DARCcfg.TimeBase!=0xFF)
	{
		// timer 4 used for PWM
		PR4 = DARCcfg.PWMPeriod;  // timer period
		T3CONbits.T3CCP1 = 1;  // set PWM to timer 4
		T3CONbits.T3CCP2 = 1;  // set PWM to timer 4
		T4CON = 0b00000100;  // 0.2us timebase
		if (DARCcfg.TimeBase==1) T4CON |= 0b00000001;  // 0.8us timebase
		if (DARCcfg.TimeBase==2) T4CON |= 0b00000010;  // 3.2us timebase
		if (DARCcfg.PinFuncCCP1==2) { DirCCP1 = DirOutput; CCP1CON = 0x0F;  CCPR1L = 0; } // turn PWM on, zero output
		if (DARCcfg.PinFuncCCP2==2) { DirCCP2 = DirOutput; CCP2CON = 0x0F;  CCPR2L = 0; } // turn PWM on, zero output
		if (DARCcfg.PinFuncCCP3==2) { DirCCP3 = DirOutput; CCP3CON = 0x0F;  CCPR3L = 0; } // turn PWM on, zero output
		if (DARCcfg.PinFuncCCP4==2) { DirCCP4 = DirOutput; CCP4CON = 0x0F;  CCPR4L = 0; } // turn PWM on, zero output
		if (DARCcfg.PinFuncCCP5==2) { DirCCP5 = DirOutput; CCP5CON = 0x0F;  CCPR5L = 0; } // turn PWM on, zero output
	}

	// digital I/O - ignored if analog
	DirAN0 = (DARCcfg.PinFuncAN0==0) ? 1 : 0;
	DirAN1 = (DARCcfg.PinFuncAN1==0) ? 1 : 0;
	DirAN2 = (DARCcfg.PinFuncAN2==0) ? 1 : 0;
	DirAN3 = (DARCcfg.PinFuncAN3==0) ? 1 : 0;
	DirAN4 = (DARCcfg.PinFuncAN4==0) ? 1 : 0;
	DirAN5 = (DARCcfg.PinFuncAN5==0) ? 1 : 0;
	DirAN6 = (DARCcfg.PinFuncAN6==0) ? 1 : 0;
	DirAN7 = (DARCcfg.PinFuncAN7==0) ? 1 : 0;
	DirAN8 = (DARCcfg.PinFuncAN8==0) ? 1 : 0;
	DirAN9 = (DARCcfg.PinFuncAN9==0) ? 1 : 0;
	DirAN10 = (DARCcfg.PinFuncAN10==0) ? 1 : 0;
	DirAN11 = (DARCcfg.PinFuncAN11==0) ? 1 : 0;
	DirCCP1 = (DARCcfg.PinFuncCCP1==0) ? 1 : 0;
	DirCCP2 = (DARCcfg.PinFuncCCP2==0) ? 1 : 0;
	DirCCP3 = (DARCcfg.PinFuncCCP3==0) ? 1 : 0;
	DirCCP4 = (DARCcfg.PinFuncCCP4==0) ? 1 : 0;
	DirCCP5 = (DARCcfg.PinFuncCCP5==0) ? 1 : 0;
	DirRxD = (DARCcfg.PinFuncRxD==0) ? 1 : 0;
	DirTxD = (DARCcfg.PinFuncTxD==0) ? 1 : 0;
	DirINT0 = (DARCcfg.PinFuncINT0==0) ? 1 : 0;
	DirINT1 = (DARCcfg.PinFuncINT1==0) ? 1 : 0;
	DirSCL = (DARCcfg.PinFuncSCL==0) ? 1 : 0;
	DirSDA = (DARCcfg.PinFuncSDA==0) ? 1 : 0;
	DirSDO = (DARCcfg.PinFuncSDO==0) ? 1 : 0;

	// make sure controls based on inputs are correct
	RefreshInputs();

	// main loop
	while ( 1 ) 
	{
		// new for V3.0.00005 - reset BlueMatik on general error
		if (ResetMePlease)
		{
			Reset();
		}

		// if PulseClearCountDown is nonzero, decrement until zero and wait 1ms each time and then clear all pulse outputs; otherwise just pause 50ms
		if (PulseClearCountDown)
		{
			while (PulseClearCountDown)
			{
				msDelay( 1 );

				PulseClearCountDown--;
				if (PulseClearCountDown==0)
				{
					// clear any pins tied to button and image controls
					char iCtl;
					for (iCtl = 0; iCtl < pDevD->nCtl; iCtl++)
					{
						if (pCtlD[iCtl].Type==TP_CtlTypeButton || pCtlD[iCtl].Type==TP_CtlTypeImage) 
							SetDigitalPin( DARCcfg.ToFromPin[pCtlD[iCtl].ID], 0 );
					}
				}
			}
		}
		else
			msDelay( 50 );

		// if 100ms, 200ms or 500ms, pause appropriate amount of time and update inputs; note that due to interrupts etc this timing is approximate
		if ( DARCcfg.RefreshRate > 0 && DARCcfg.RefreshRate < 4)
		{
			if ( DARCcfg.RefreshRate==1 ) msDelay( 50 );   // 100 ms
			if ( DARCcfg.RefreshRate==2 ) msDelay( 150 );  // 200 ms
			if ( DARCcfg.RefreshRate==3 ) 				   // 500 ms
			{
				msDelay( 200 );
				msDelay( 250 );
			}
			RefreshInputs();
		}

		// otherwise if RefeshInputsNow semaphore raised, update inputs
		if (RefeshInputsNow)
		{
			RefreshInputs();
			RefeshInputsNow = 0;
		}



	}  // end of while ( 1 ) 

}

// HighInterrupt handler - nothing needs to be done
void HighInterrupt (void)
{
}

// LowInterrupt handler 
void LowInterrupt (void)
{
	if (IsSWI( SWI_Tick ) )
	{
		// Flash LED to show we're alive
		LedGreen = ~LedGreen;
		ClearSWI( SWI_Tick );		// Clear clock interrupt flag

		// if time to update inputs, raise RefeshInputsNow semaphore
		if ( DARCcfg.RefreshRate == 4 ) // every second 
		{
			RefeshInputsNow = 0xFF;
			return;
		}
		if ( DARCcfg.RefreshRate == 5 && (RealTimeClock.sec & 0x01) == 0 )  // every 2 seconds
		{
			RefeshInputsNow = 0xFF;
			return;
		}
		if ( DARCcfg.RefreshRate == 6 && (RealTimeClock.sec % 5) == 0 )  // every 5 seconds
		{
			RefeshInputsNow = 0xFF;
			return;
		}
		if ( DARCcfg.RefreshRate == 7 && (RealTimeClock.sec % 10) == 0 )  // every 10 seconds
		{
			RefeshInputsNow = 0xFF;
			return;
		}
		if ( DARCcfg.RefreshRate == 8 && (RealTimeClock.sec % 15) == 0 )  // every 15 seconds
		{
			RefeshInputsNow = 0xFF;
			return;
		}
		if ( DARCcfg.RefreshRate == 8 && (RealTimeClock.sec % 30) == 0 )  // every 30 seconds
		{
			RefeshInputsNow = 0xFF;
			return;
		}

		// if not at the top of the minute, return
		if (RealTimeClock.sec!=0) return;

		if ( DARCcfg.RefreshRate == 9 )  // every minute
		{
			RefeshInputsNow = 0xFF;
			return;
		}
		if ( DARCcfg.RefreshRate == 10 && (RealTimeClock.min & 0x01) == 0 )  // every 2 minutes
		{
			RefeshInputsNow = 0xFF;
			return;
		}
		if ( DARCcfg.RefreshRate == 11 && (RealTimeClock.min % 5) == 0 )     // every 5 minutes
		{
			RefeshInputsNow = 0xFF;
			return;
		}
		if ( DARCcfg.RefreshRate == 12 && (RealTimeClock.min % 10) == 0 )     // every 10 minutes
		{
			RefeshInputsNow = 0xFF;
			return;
		}
		if ( DARCcfg.RefreshRate == 13 && (RealTimeClock.min % 15) == 0 )     // every 15 minutes
		{
			RefeshInputsNow = 0xFF;
			return;
		}
		if ( DARCcfg.RefreshRate == 14 && (RealTimeClock.min % 30) == 0 )     // every 30 minutes
		{
			RefeshInputsNow = 0xFF;
			return;
		}
		if ( DARCcfg.RefreshRate == 15 && RealTimeClock.min == 0 )  		   // every 60 minutes
		{
			RefeshInputsNow = 0xFF;
			return;
		}

		return;
	}
}


// Error event handler 
void ErrorStatus( unsigned char ErrNum )
{
	if (DARCcfg.Initialized!=0x55 || (!DARCcfg.ResetOnError))
	{
		// Flash error number for diagnositic purposes
		Breakpoint( ErrNum );
	}

	//  Now reset in any event
	Reset();
}


// BlueMatik Event handler 
void LMTEvent( unsigned char EventID, void *pData1, void *pData2 )
{
	if ( EventID==LMTE_Syntax )
	{
		if (((unsigned char*)pData1)[0] == 0x0F) // general error
		{
			ResetMePlease = True;
			return;
		}

		if (DARCcfg.Initialized!=0x55 || (!DARCcfg.ResetOnError))
		{
			// Flash error number for diagnositic purposes
			Breakpoint( ((unsigned char*)pData1)[0] );
		}

		//  Now reset in any event
		Reset();
	}
}

// FxPEvent handler 
void FxPEvent( unsigned char EventID, void *pData )
{
	if ( EventID==FxPE_Connect ) 
	{
		// turn off red led during connection
		LedRed = LedRedOn;
	}

	if ( EventID==FxPE_Disco ) 
	{
		// turn off red led after disconnection 
		LedRed = LedRedOff;
	}

	if ( EventID==FxPE_ClntUpdate ) 
	{
		// process control
		ControlChanged( * ((unsigned short*) pData) ); 
	}

	// As of V3.0.00002, ping fail can unexpectedly fail so it should be turned off 
	// this legacy code generates a '222' breakpoint if a ping fail error occurs
	if ( EventID==FxPE_PingFail )
	{
		if (DARCcfg.Initialized!=0x55 || (!DARCcfg.ResetOnError))
		{
			// Flash indication of error for diagnositic purposes
			Breakpoint( 222 );
		}

		//  Now reset in any event
		Reset();
	}
}

void ControlChanged( unsigned char CtlID )
{
	// identify control index
	unsigned char iCtlIdx;
	for (iCtlIdx = 0; iCtlIdx < pDevD->nCtl; iCtlIdx++)
	{
		if (pCtlD[ iCtlIdx ].ID == CtlID) break;
	}

	// if can't find control, return
	if (iCtlIdx==pDevD->nCtl) return;

	// ignore if control not connected to a pin
	if (DARCcfg.ToFromPin[CtlID]==0) return;

	switch ( (unsigned char)pCtlD[ iCtlIdx ].Type )
	{
		case TP_CtlTypeImage:  
		case TP_CtlTypeButton:  
		{
			// if tied to an output pin, set high
			SetDigitalPin( DARCcfg.ToFromPin[CtlID], 1 );

			// tell main loop to clear all pulse pins after 50ms
			PulseClearCountDown = 50;
			return;
		}
		case TP_CtlTypeLatch:
		{
			// if tied to an output pin, set to latch value
			char iVal;
			GetBytes( pCtlD[ iCtlIdx ].VarStr, pCtlD[ iCtlIdx ].VarLoc, &iVal, 1  );
			if (iVal)
				SetDigitalPin( DARCcfg.ToFromPin[CtlID], 1 );
			else
				SetDigitalPin( DARCcfg.ToFromPin[CtlID], 0 );
			return;
		}
		case TP_CtlTypeList:
		{
			short sVal;
			GetBytes( pCtlD[ iCtlIdx ].VarStr, pCtlD[ iCtlIdx ].VarLoc, &sVal, 2  );
			SetListPin( DARCcfg.ToFromPin[CtlID], sVal );
			return;
		}
		case TP_CtlTypeNumber:
		{
			short sVal;
			GetBytes( pCtlD[ iCtlIdx ].VarStr, pCtlD[ iCtlIdx ].VarLoc, &sVal, 2  );
			SetNumberPin( DARCcfg.ToFromPin[CtlID], sVal );
			return;
		}
	}
}

void SetDigitalPin( unsigned char DigitalPinID, unsigned char iVal )
{
	switch (DigitalPinID)
	{
		case 1: AN0Pin = iVal; return;
		case 2: AN1Pin = iVal; return;
		case 3: AN2Pin = iVal; return;
		case 4: AN3Pin = iVal; return;
		case 5: AN4Pin = iVal; return;
		case 6: AN5Pin = iVal; return;
		case 7: AN6Pin = iVal; return;
		case 8: AN7Pin = iVal; return;
		case 9: AN8Pin = iVal; return;
		case 10: AN9Pin = iVal; return;
		case 11: AN10Pin = iVal; return;
		case 12: AN11Pin = iVal; return;
		case 13: CCP1Pin = iVal; return;
		case 14: CCP2Pin = iVal; return;
		case 15: CCP3Pin = iVal; return;
		case 16: CCP4Pin = iVal; return;
		case 17: CCP5Pin = iVal; return;
		case 18: INT0Pin = iVal; return;
		case 19: INT1Pin = iVal; return;
		case 20: TxDPin = iVal; return;
		case 21: RxDPin = iVal; return;
		case 22: SDAPin = iVal; return;
		case 23: SCLPin = iVal; return;
		case 24: SDOPin = iVal; return;
	}
}

void SetListPin( unsigned char DigitalPinID, unsigned short iVal )
{
	SetNumberPin( DigitalPinID+12, iVal );
}

void SetNumberPin( unsigned char DigitalPinID, unsigned short iVal )
{
	switch (DigitalPinID)
	{
		case 13: // Parallel A
		{
			switch (DARCcfg.ParallelA)
			{
				case 2: SetPA2( iVal ); return;
				case 3: SetPA3( iVal ); return;
				case 4: SetPA4( iVal ); return;
				case 5: SetPA5( iVal ); return;
				case 6: SetPA6( iVal ); return;
				case 7: SetPA7( iVal ); return;
			}
			return;
		}
		case 14: // Parallel B
		{
			switch (DARCcfg.ParallelB)
			{
				case 2: SetPB2( iVal ); return;
				case 3: SetPB3( iVal ); return;
				case 4: SetPB4( iVal ); return;
				case 5: SetPB5( iVal ); return;
			}
			return;
		}
		case 15: // Parallel C
		{
			switch (DARCcfg.ParallelC)
			{
				case 2: SetPC2( iVal ); return;
				case 3: SetPC3( iVal ); return;
				case 4: SetPC4( iVal ); return;
				case 5: SetPC5( iVal ); return;
			}
			return;
		}
		case 16: // CCP1 PWM
		{
			// lower 2 bits
			CCP1CON = (( ((unsigned char)iVal)&0b00000011) << 4) | 0b00001100;
			// upper 8 bits
			iVal = iVal >> 2;
			CCPR1L = (unsigned char)iVal;
			return;
		}
		case 17: // CCP2 PWM
		{
			// lower 2 bits
			CCP2CON = (( ((unsigned char)iVal)&0b00000011) << 4) | 0b00001100;
			// upper 8 bits
			iVal = iVal >> 2;
			CCPR2L = (unsigned char)iVal;
			return;
		}
		case 18: // CCP3 PWM
		{
			// lower 2 bits
			CCP3CON = (( ((unsigned char)iVal)&0b00000011) << 4) | 0b00001100;
			// upper 8 bits
			iVal = iVal >> 2;
			CCPR3L = (unsigned char)iVal;
			return;
		}
		case 19: // CCP4 PWM
		{
			// lower 2 bits
			CCP4CON = (( ((unsigned char)iVal)&0b00000011) << 4) | 0b00001100;
			// upper 8 bits
			iVal = iVal >> 2;
			CCPR4L = (unsigned char)iVal;
			return;
		}
		case 20: // CCP5 PWM
		{
			// lower 2 bits
			CCP5CON = (( ((unsigned char)iVal)&0b00000011) << 4) | 0b00001100;
			// upper 8 bits
			iVal = iVal >> 2;
			CCPR5L = (unsigned char)iVal;
			return;
		}
	}
}


void RefreshInputs( void )
{
	unsigned char NumUpdate = 0;
	unsigned char iCtlIdx; 

	DisableLowInts;
	
	// loop through all controls 
	for (iCtlIdx = 0; iCtlIdx < pDevD->nCtl; iCtlIdx++)
	{
		unsigned char ToFromPin ;
		rom bqFxPCtlData* pCtl = &pCtlD[ iCtlIdx ];

		ToFromPin = DARCcfg.ToFromPin[ pCtl->ID ];

		// ignore if control not connected to a pin
		if (ToFromPin==0) continue;
	
		switch ( (unsigned char)pCtl->Type )
		{
			case TP_CtlTypeText:
			{
				unsigned char TextStart;
				unsigned char HalfTextLen;
				unsigned char charCount;
				unsigned char ch;

				// identify boundaryu between 'OFF' and 'ON' strings
				// assumes VarLoc is exactly two equal length strings concatenated followed by zero terminator
				// assumes VarLen is exactly length to zero terminator
				TextStart = 0;
				if (pDevD->ServerFlags&MTF_Unicode) 
					HalfTextLen = (pCtl->VarLen-2) / 2 ;
				else
					HalfTextLen = (pCtl->VarLen-1) / 2 ;

				// get value and set text accordingly;
				if (GetPinVal( ToFromPin )) TextStart +=HalfTextLen;

				// set text
				for (charCount = 0; charCount <=HalfTextLen + ((pDevD->ServerFlags&MTF_Unicode) ? 1 : 0); charCount++)
				{
					if (charCount>=HalfTextLen)
						ch = 0;  // zero terminator(s)
					else
						GetBytes( pCtl->IniStr, pCtl->IniLoc+TextStart+charCount, &ch, 1 );
					SetBytes( pCtl->VarStr, pCtl->VarLoc+charCount, 0, &ch, 1 );
				}

				// update client
				NumUpdate ++;
				continue;
			}
			case TP_CtlTypeNumber:
			{
				short sVal;

				// ignore if modifiable
				if (pCtl->dwflags & CTL_NUM_MODIFIABLE) continue;

				// get value
				sVal = GetNumberVal( ToFromPin );		

				// set control
				SetBytes( pCtl->VarStr, pCtl->VarLoc, 0, &sVal, sizeof(short) )	;
				NumUpdate ++;
				continue;
			}
			case TP_CtlTypeMatrix:  // assumes it's a date-time matrix and real time clock is working
			{
				unsigned short sVal;
				long RowToUpdate;
				long NextRowCounter;
				long RowCounter;
				short NumRows;
				unsigned char  NumCols;
				unsigned char  CellSize;
				unsigned char  RowSize;
				ADD Addr;

				// retrieve value
				sVal = GetNumberVal( ToFromPin );	

				// calculate matrix dimensions
				CellSize = 1;
				if ( (pCtl->dwflags & CTL_MTX_Y_BYTE_MASK) == CTL_MTX_Y_2BYTE ) CellSize = 2;
				if ( (pCtl->dwflags & CTL_MTX_Y_BYTE_MASK) == CTL_MTX_Y_4BYTE ) CellSize = 4;
				NumCols = pCtl->nCtlVal;

				// get RowCounter value - always last 4 bytes of matrix data
				GetBytes( pCtl->VarStr, pCtl->VarLoc + pCtl->VarLen - 4, &RowCounter, 4  );

				// Row size in bytes is CellSize * NumCols plus 8 bytes for Date Time data
				RowSize = CellSize * NumCols + sizeof( DateTimeU );
				NumRows = (pCtl->VarLen - 4) / RowSize;

				// calculate which row will be updated
				RowToUpdate = ( (RowCounter>=0) ? ( (RowCounter< NumRows) ? RowCounter : 0 ) : ( (RowCounter> -NumRows) ? - RowCounter : 0 ) );

				// calculate address of cell to update (put data in first column)
				Addr = pCtl->VarLoc + RowToUpdate * CellSize;

				// set new value
				SetBytes( pCtl->VarStr, Addr, 0, &sVal, (CellSize==4) ? 2 : CellSize );

				// calculate address of datetime cell to update - datetime data comes after all cell data
				Addr = pCtl->VarLoc + NumCols * NumRows * CellSize + RowToUpdate * sizeof( DateTimeU );

				// set new value
				SetBytes( pCtl->VarStr, Addr, 0, (unsigned char *)&RealTimeClock, sizeof( DateTimeU ) );

				// calculate next row counter value
				NextRowCounter = ( (RowCounter>=0) ? ( (RowCounter< NumRows) ? RowCounter+1 : -1 ) : ( (RowCounter> -NumRows) ? RowCounter - 1 : -1 ) );

				// set RowCounter value - always last 4 bytes of matrix data
				SetBytes( pCtl->VarStr, pCtl->VarLoc + pCtl->VarLen - 4, 0, &NextRowCounter, 4  );

				// send part update message
		      	FxPCommand( FxPC_PartUpdate, pCtl->ID, (void *) RowToUpdate );
				continue;
			}
		}
	}


	// loop through all controls and send update message
	if (NumUpdate)
	{

		// start update messsage
		FxPCommand( FxPC_MultiUpdate, NumUpdate, 0 );

		for (iCtlIdx = 0; iCtlIdx < pDevD->nCtl; iCtlIdx++)
		{
			unsigned char ToFromPin ;
			rom bqFxPCtlData* pCtl = &pCtlD[ iCtlIdx ];
		
			ToFromPin = DARCcfg.ToFromPin[ pCtl->ID ];
		
			// ignore if control not connected to a pin
			if (ToFromPin==0) continue;
		
			switch ( (unsigned char)pCtl->Type )
			{
				case TP_CtlTypeNumber:
					if (pCtl->dwflags & CTL_NUM_MODIFIABLE) continue;
					// fall through

				case TP_CtlTypeText:
					// ignore if modifiable
					if (pCtl->dwflags & CTL_NUM_MODIFIABLE) continue;
					FxPCommand( FxPC_CtlUpdate, pCtl->ID, 0 );
					continue;		
			}
		}
	}

	EnableLowInts;

}


unsigned char GetPinVal( unsigned char DigitalPinID )
{
	switch (DigitalPinID)
	{
		case 1: return AN0Pin; 
		case 2: return AN1Pin;
		case 3: return AN2Pin;
		case 4: return AN3Pin;
		case 5: return AN4Pin;
		case 6: return AN5Pin;
		case 7: return AN6Pin;
		case 8: return AN7Pin;
		case 9: return AN8Pin;
		case 10: return AN9Pin;
		case 11: return AN10Pin;
		case 12: return AN11Pin;
		case 13: return CCP1Pin;
		case 14: return CCP2Pin;
		case 15: return CCP3Pin;
		case 16: return CCP4Pin;
		case 17: return CCP5Pin;
		case 18: return INT0Pin;
		case 19: return INT1Pin;
		case 20: return TxDPin;
		case 21: return RxDPin;
		case 22: return SDAPin;
		case 23: return SCLPin;
		case 24: return SDOPin;
	}
}

short GetNumberVal( unsigned char PinID )
{
	// analog
	if (PinID>0 && PinID<13)
	{
		unsigned short ADResult;
		SetADChan( PinID-1 );  					// select a to d channel
		CyclesDelay3p2plus3p2times( 4 ); 		// 16us delay to charge S & H cap (refer to PIC18F6720 manual)
		StartAtoD;								// start a to d conversion
		AwaitAtoDComplete;						// await end of conversion
		GetADResult10bit( ADResult );			// get result
		return ADResult;
	}

	// parallel
	switch (PinID)
	{
		case 13: // Parallel A
		{
			switch (DARCcfg.ParallelA)
			{
				case 2: return GetPA2;
				case 3: return GetPA3;
				case 4: return GetPA4;
				case 5: return GetPA5;
				case 6: return GetPA6;
				case 7: return GetPA7;
			}
			break;
		}
		case 14: // Parallel B
		{
			switch (DARCcfg.ParallelB)
			{
				case 2: return GetPB2;
				case 3: return GetPB3;
				case 4: return GetPB4;
				case 5: return GetPB5;
			}
			break;
		}
		case 15: // Parallel C
		{
			switch (DARCcfg.ParallelC)
			{
				case 2: return GetPC2;
				case 3: return GetPC3;
				case 4: return GetPC4;
				case 5: return GetPC5;
			}
			break;
		}
	}
	return 0;
}
